% The script to visualize the admissible input sets for the quadrotor example.
% 
% S. Kaynama (kaynama@ece.ubc.ca), last modified 12/10/14

clear all

gravity = 9.81;
uhover = 8.96; %Thrust value needed for hover (empirically measured); see Kene's email (flagged)

kT = gravity/(uhover^2);
kp = 1/0.15;

xeq = [2 0 sqrt(gravity/kT)]';
ueq = sqrt(gravity/kT); %(The same as uhover.)

a = [0 1 0;
    0 0 2*kT*xeq(3);
    0 0 -kp];

b = [0 0 kp]';

g = [0 1 0]'; 

n = size(a,1);
m = size(b,2);

%Control set \U
uMin = 0;
uMax = 10;
uLinMax = uMax-ueq;
uLinMin = uMin-ueq;
U = (uLinMax-uLinMin)/2*ell_unitball(m)+(uLinMax+uLinMin)/2;

%Disturbance set \V
vMax = .1; %Affects x2 only (in +ve direction), depends on value of x3 { CHANGE }
vMin = 0;
V = ellipsoid((vMax+vMin)/2*ones(1,1),((vMax-vMin)/2)^2*eye(1));

%Constraint set \K
ceiling = 2.8; floor = 0.5; %meters
acclMin = 8; acclMax = 10; %bounds on acceleration x3, determines disturbance on x2 { CHANGE } 
velMin = -1.5; velMax = 1.5;
K0 = ellipsoid(-xeq + [(ceiling+floor)/2; (velMin+velMax)/2 ;(acclMin+acclMax)/2],...
    diag([ ((ceiling-floor)/2)^2 ((velMax-velMin)/2)^2 ((acclMax-acclMin)/2)^2 ]));

d = 1/10; %Sampling time at 10-30 Hz.
T = [0 1]; %Horizon.
N = T(2)/d;
InvRes = d/200; %Sub-sample step length (i.e. resolution in Inv calculation)

Inm = eye(n+m);
[Uq, UQ] = double(U);

%Compute the sets 
method = 2; %Always recommended to be 2 (method 1 is unnecessarily overly conservative).
L0_num = 1;
ignoreUinMaxVol = 0; %1 may yield smaller control sets.
[K_all, Inv_all] = QuadrotorSDEllipsoidalKernels_v1(method,L0_num,0,ignoreUinMaxVol);

l = 1; %For now just do things for 1 direction.
Disc = K_all{l}(end-1); %Only do things for Disc_T and Inv_T.
Inv = Inv_all{l}(end-1);



%% 


ng = 20;
x1 = linspace(floor-xeq(1),ceiling-xeq(1),ng); 
x2 = linspace(velMin-xeq(2),velMax-xeq(2),ng);
x3 = linspace(acclMin-xeq(3),acclMax-xeq(3),ng);


%For shaded regions
fillcolor = [230,230,250]/255;% [255,228,225]/255;
xfill = -10:10;
Xfill=[xfill,fliplr(xfill)];                

yfillup1=(Uq+sqrt(UQ))*ones(size(xfill));
yfillup2=10*(Uq+sqrt(UQ))*ones(size(xfill));
Yfillup=[yfillup1,fliplr(yfillup2)];              

yfillbo1=(Uq-sqrt(UQ))*ones(size(xfill));
yfillbo2=10*(Uq-sqrt(UQ))*ones(size(xfill));
Yfillbo=[yfillbo1,fliplr(yfillbo2)];              

rows = 2;
cols = 4;

figure
%Final state dimension, position:
count = 0;
for i = 1:2:ng,
    UfbFixedx3 = Inv;
    H = hyperplane(Inm(:,3),x3(i));
    UfbFixedx3 = hpintersection(UfbFixedx3,H);
    
    if ~isempty(UfbFixedx3) && ~isempty(ellintersection_ia([U projection(UfbFixedx3,Inm(:,4))])),
        
        %Set up values of x2 at which we slice
        UfbFixedx3Pr = projection(UfbFixedx3,Inm(:,[1 2 4]));
        [UfbFixedx3Prq, UfbFixedx3PrQ] = double(UfbFixedx3Pr);
        [v,e] = eig(UfbFixedx3PrQ);
        e = sqrt(e); %Length of major and minor axes.
        e(abs(e)<1e-8)=0;
        ve = v*e; %Major and minor axes for ellipsoid with fixed x3.
        [~,supp] = rho(UfbFixedx3Pr,[ve -ve]);
        alpha = .8;
        x2new(1) = alpha*min(supp(2,:)) + (1-alpha)*UfbFixedx3Prq(2);
        x2new(2) = UfbFixedx3Prq(2);
        x2new(3) = x2new(2) + abs(x2new(2)-x2new(1));  %(1-alpha)*UfbFixedx3Prq(2) + alpha*max(supp(2,:));
                
        count = count+1;
        if count > rows*cols, break; end
        subplot(rows,cols,count)
        hold on, grid on
        
        for k = 1:3,
            Ufb = UfbFixedx3;
            %H = hyperplane(Inm(:,2),x2(k));
            H = hyperplane(Inm(:,2),x2new(k));
            Ufb = hpintersection(Ufb,H);
            if ~isempty(Ufb),
                Ufb = projection(Ufb,Inm(:,[1 4]));
                plot(Ufb,'b')
                [Ufbq,~] = double(Ufb);
                text(Ufbq(1)+0.005,Ufbq(2)+0.005, sprintf('\\leftarrow %2.2f',x2new(k)),...
                    'HorizontalAlignment','left')
            end
        end
        xlim([floor-xeq(1) ceiling-xeq(1)])
        ylim([-1.8 1.8])
        xyaxis = axis;
        hfill = fill(Xfill,Yfillup,fillcolor); 
        uistack(hfill, 'bottom')
        hfill = fill(Xfill,Yfillbo,fillcolor); 
        uistack(hfill, 'bottom')
        axis(xyaxis);
        xlabel('\bar{x}_1'), ylabel('\bar{u}')
        title(sprintf('\\bar{x}_3 = %2.2f', x3(i)))
    end
end


ax = axes('position',[0,0,1,1],'visible','off'); 
tx = text(0.44,0.98,sprintf('Ignore u-space for maxvol? %2.0f', ignoreUinMaxVol)); 
set(tx,'fontweight','bold');

