function [E, S] = ellIntersectionIAProjET_tradeoff(EE,Proj1,Proj2,alpha)
%
% [E, S] = ellIntersectionIAProjET_tradeoff(EE,Proj1,Proj2,alpha)
%
% Computes the ellipsoid in the intersection of two or more ellipsoids 
% such that the weighted volume of its two projections is maximized.
% 
% Proj1 and Proj2 must be the non-symmetric matrices such that
% Proji'*Proji = P is a projection matrix, i.e. P^2=P. E.g.: Proj1=[1 0 0;
% 0 1 0] and Proj2=[0 0 1]. Alpha is the weight of the projection onto the
% second subspace.
%
% Adapted from ELLIPSOID/ELLINTERSECTION_IA by Alex Kurzhanskiy.
% Shahab Kaynama, Oct 2014, kaynama@ece.ubc.ca


global ellOptions;

if ~isstruct(ellOptions)
    evalin('base', 'ellipsoids_init;');
end

dims = dimension(EE);
mn   = min(min(dims));
mx   = max(max(dims));

if mn ~= mx
    error('ELLINTERSECTION_IA: all ellipsoids must be of the same dimension.');
end

[m, n] = size(EE);
M      = m * n;
EE     = reshape(EE, 1, M);
zz     = zeros(mn, 1);
I      = eye(mn);

if ellOptions.verbose > 0
    fprintf('Invoking YALMIP...\n');
end

B      = sdpvar(mn, mn);
d      = sdpvar(mn, 1);
ll     = sdpvar(M, 1);


cnstr = lmi;

for i = 1:M
    [q, Q] = double(EE(i));
    
    if rank(Q) < mn
        Q = regularizeET(Q); %Modified
    end
    
    A     = ell_inv(Q);
    b     = -A * q;
    c     = q' * A * q - 1;
    X     = [(-ll(i,1)-c+b'*Q*b) zz'       (d+Q*b)'
        zz                  ll(i,1)*I B
        (d+Q*b)             B         Q]; %#ok<NASGU>
    
    cnstr = cnstr + set('X>=0');
    cnstr = cnstr + set('ll(i, 1)>=0');
end


f = -(1-alpha)*trace(Proj1*B*Proj1')-alpha*trace(Proj2*B*Proj2');
% f = -logdet(B);

% f = -(1-alpha)*trace(Proj1*B*Proj1')-alpha*trace(Proj2*B*Proj2')-trace(B);
% f = -(1-alpha)*trace(Proj1*B*Proj1')-alpha*trace(B);

S = solvesdp(cnstr, f, ellOptions.sdpsettings); %Modified.
B = double(B);
d = double(d);

if rank(B) < mn
    B = regularizeET(B); %Modified
end

P = B * B';
P = (1-ellOptions.abs_tol)*0.5*(P + P');

E = ellipsoid(d, P);

if nargout < 2
    clear S;
end

return;
